/*
 * dm9pci.c
 * Copyright (c) 2000-2002 DAVICOM Semiconductor Inc. All Rights Reserved 
 *
 * This file implement the QNXRTP driver for DAVICOM DM9PCI Series Fast
 * Ethernet adapter.
 *
 * Author: John H.T. Ho (john_ho@davicom.com.tw)
 *
 * Modification History
 *
 * Date	      Modfier.	 Modification Descriptions
 * ---------- ---------- ------------------------------------------
 * 2000.10.23  htho	 Modify based on the SMC driver from QSSL
 * 2002.1.4    htho	 Report line status for NICINFO 	 	
 */
#include <dm9pci.h>


#define NIC_INTERRUPT_EVENT		0
#define NIC_TIMER_EVENT			1
#define NIC_PRIORITY			21

/************************************************************************
 *
 *			Private Routines Declaration
 *
 ************************************************************************/
void dump_tx_desc(Nic_Dm9pci_Ext_t *ext);
void dump_rx_desc(Nic_Dm9pci_Ext_t *ext);
/*
 * MII PHY Functions
 */
static void phy_write(Nic_Dm9pci_Ext_t *ext, uint8_t phy_addr, uint8_t offset, uint16_t phy_data);
static uint16_t phy_read(Nic_Dm9pci_Ext_t *ext, uint8_t phy_addr, uint8_t offset);
static void phy_write_1bit(Nic_Dm9pci_Ext_t *ext, uint32_t phy_data);
static uint16_t phy_read_1bit(Nic_Dm9pci_Ext_t *ext);
/*
 * SROM Functions
 */
static uint16_t read_srom_word(Nic_Dm9pci_Ext_t *ext, int offset);
/*
 * timer interrupt service routine
 */
#if TIMER_ACTIVE
int dm9pci_timer( Nic_t *nic );
void dynamic_auto_detect(Nic_t *nic);
#endif
/*
 * interrupt service routine
 */
//static int32 dm9pci_interrupt(void *_ext);

/*
 * Tx/Rx descriptor functions
 */
int dm9pci_descriptor_init(Nic_t *nic);
void dm9pci_descriptor_free(Nic_Dm9pci_Ext_t *ext);

static void init_HPNA_PHY(Nic_Dm9pci_Ext_t *ext);
static void init_DM9102AF_E3(Nic_Dm9pci_Ext_t *ext);
static void local_var_init(Nic_Dm9pci_Ext_t *ext);
static void set_PHY_ANAR(Nic_Dm9pci_Ext_t *ext);
static void startup_auto_detect(Nic_t *nic);
static void auto_detect_media_type(Nic_t *nic);
static void set_CSR6_op_mode(Nic_Dm9pci_Ext_t *ext);
static void update_cr6(Nic_Dm9pci_Ext_t *ext, uint32_t cr6_data);
static int send_setup_frame(Nic_Dm9pci_Ext_t *ext);

/************************************************************************
 *
 *			QNXRTP Network Driver Entry Data Structure
 *
 ************************************************************************/
io_net_dll_entry_t io_net_dll_entry = {
		2, 
		dm9pci_init, 
		NULL };

/************************************************************************
 *
 *			Functions in our driver that IO-NET can call
 *
 ************************************************************************/
io_net_registrant_funcs_t dm9pci_funcs = {
		8, 
		NULL, 
		dm9pci_send_packets,
		dm9pci_receive_complete, 
		dm9pci_shutdown1,
		dm9pci_shutdown2,
		dm9pci_advertise,
		dm9pci_devctl, 
		dm9pci_flush, 
		NULL
};

/************************************************************************
 *
 *			Information used when registering with IO-NET
 *
 ************************************************************************/
io_net_registrant_t dm9pci_entry = {
        _REG_PRODUCER_UP,
        "devn-dm9pci",
        "en",
        NULL,
        NULL,
        &dm9pci_funcs,
        0
	};

/************************************************************************
 *
 *			Private Data Declarations
 *
 ************************************************************************/
#if CONSOLE_DEBUG
FILE *dfd;
#endif
/************************************************************************
 *
 *			Driver Entry Points
 *
 ************************************************************************/
int dm9pci_init( void *dll_hdl, dispatch_t *dpp, io_net_self_t *ion, char *options)
{
	dpp = dpp;
	setbuf(stdout, NULL);

#if CONSOLE_DEBUG
dfd = fopen("/dev/ser1", "w");
#endif
	return( dm9pci_detect( dll_hdl, ion, options ) );
}

int dm9pci_reset( Nic_t *nic )
{
	Nic_Dm9pci_Ext_t *ext;

#if DEBUG_DETECT
fprintf(dfd, "dm9pci_reset()\n\r");
#endif
	ext	= (Nic_Dm9pci_Ext_t *)nic->ext;

	/* local variables init */
	local_var_init(ext);

	/* Reset DM9PCI MAC: need 32 PCI clock to complete */
	OUT32(ext, CSR0, CSR0_SR);	/* reset MAC */
	DELAY_5US;
	OUT32(ext, CSR0, ext->cr0_data);

	/* Initiliaze Transmit/Receive decriptor and CSR3/CSR4 */
	if( dm9pci_descriptor_init(nic) )
		return(1);
//dump_tx_desc(ext);
//dump_rx_desc(ext);

	/* Send setup frame */
	if( send_setup_frame(ext) )
		return(1);
 
	/* Init CSR5, CSR7, interrupt active bit */
	OUT32(ext, CSR5, 0xffffffff);	/* clear all CR5 status */
	OUT32(ext, CSR7, ext->cr7_data);

	/* Init CR15, Tx jabber and Rx watchdog timer */
	OUT32(ext, CSR15, ext->cr15_data);
	
	/* set CR6 according to the operating mode of PHY */
	set_CSR6_op_mode(ext);

	/* Enable DM9PCI MAC Tx/Rx function */
	ext->cr6_data |= (CSR6_RXSC|CSR6_TXSC);
	update_cr6(ext, ext->cr6_data);
	
	ext->open = 1;
	
	return( 0 );
}


int dm9pci_init_memory( Nic_t *nic )
{
	Nic_Dm9pci_Ext_t	*ext;

#if DEBUG_DETECT
fprintf(dfd, "dm9pci_init_memory()\n\r");
#endif
	ext = (Nic_Dm9pci_Ext_t *)nic->ext;

	/* allocate rx descriptors pool */
	if( ( ext->rfd = mmap( NULL, sizeof(dm9pci_rx_desc)*RX_DESC_NUM+31, PROT_READ | PROT_WRITE | PROT_NOCACHE, MAP_ANON | MAP_PHYS, NOFD, 0 ) ) == MAP_FAILED )
		return( 1 );

	/* allocate rx npkt pool */
	if( ( ext->rx_pktq = calloc( RX_DESC_NUM, sizeof( ext->rx_pktq ) ) ) == NULL )
		return( 1 );

	/* allocate tx descriptors pool */
	if( ( ext->tfd = mmap( NULL, sizeof(dm9pci_tx_desc)*TX_DESC_NUM+31, PROT_READ | PROT_WRITE | PROT_NOCACHE, MAP_ANON | MAP_PHYS, NOFD, 0 ) ) == MAP_FAILED )
		return( 1 );

	/* allocate tx npkt pool */
	if( ( ext->tx_pktq = calloc( TX_DESC_NUM, sizeof( ext->tx_pktq ) ) ) == NULL )
		return( 1 );

	/* allocate tx data buffer pool */
	if( ( ext->alloc_tx_buf = mmap( NULL, BUFFER_SIZE*TX_DESC_NUM+31, PROT_READ | PROT_WRITE | PROT_NOCACHE, MAP_ANON | MAP_PHYS, NOFD, 0 ) ) == MAP_FAILED )
		return( 1 );

	return( 0 );
}

int dm9pci_config( Nic_t *nic )
{
	int	i;
	Nic_Dm9pci_Ext_t *ext;

#if DEBUG_DETECT
fprintf(dfd, "dm9pci_config()\n\r");
#endif
	ext = (Nic_Dm9pci_Ext_t *)nic->ext;

	ext->iobase = mmap_device_io( 256, nic->cfg.IOPort_Base[0] );

	strcpy( nic->cfg.Description, "DAVICOM DM9PCI" );
	nic->media = NIC_MEDIA_802_3;
	nic->mac_length	= ETH_MAC_LEN;
	if( nic->mtu == 0 || nic->mtu > ETH_MAX_PKT_LEN )
	{
		nic->mtu = ETH_MAX_PKT_LEN;
	}

	/* read SROM */
	for (i=0; i<64; i++)
		((uint16_t *)ext->srom)[i]=read_srom_word(ext, i);

	/* Set Node address */
//fprintf(dfd, "Node Address =");
	for (i = 0; i < ETHER_ADDR_LEN; i++)
{
		nic->current_address[i] = nic->permanent_address[i] = ext->macaddr.ebyte[i] = ext->srom[20+i];
//fprintf(dfd, " %02X", ext->macaddr.ebyte[i]);
}
//fprintf(dfd, "\n\r");

	if( dm9pci_init_memory( nic ) )
		return( 1 );

	ext->cfg_mode = MEDIA_TYPE_AUTO;
	startup_auto_detect( nic );
	dm9pci_reset( nic );

	if( ext->verbose ) {
		nic_display_config( nic );
	}

	return( 0 );
}

int dm9pci_register_device( Nic_t *nic, io_net_self_t *ion, void *dll_hdl)
{
	Nic_Dm9pci_Ext_t		*ext;
	pthread_attr_t		pattr;
	pthread_mutexattr_t     mattr;
	struct sched_param	param;
	struct sigevent		event;
	uint16_t		lan;

#if DEBUG_DETECT
fprintf(dfd, "dm9pci_register_device()\n\r");
#endif
	ext = (Nic_Dm9pci_Ext_t	*)nic->ext;

	ext->ion = ion;
	ext->dll_hdl = dll_hdl;

	if( ( ext->chid = ChannelCreate( _NTO_CHF_DISCONNECT | _NTO_CHF_UNBLOCK ) ) == -1 ||
	   ( ext->coid = ConnectAttach( 0, 0, ext->chid, 0, _NTO_SIDE_CHANNEL ) ) == -1 ) {
		fprintf( stderr, "dm9pci_register_device:  Unable to attach channel and connection\n" );
		return( 1 );
	}

	pthread_mutexattr_init(&mattr);
	mattr.flags = PTHREAD_RECURSIVE_ENABLE;
	if( pthread_mutex_init( &ext->mutex, NULL ) == -1 ) {
		fprintf( stderr, "dm9pci_register_device:  Unable to initialize mutex\n" );
		return( 1 );
	}

	pthread_attr_init( &pattr );
	pthread_attr_setschedpolicy( &pattr, SCHED_RR );
	param.sched_priority = 21;
	pthread_attr_setschedparam( &pattr, &param );
	pthread_attr_setinheritsched( &pattr, PTHREAD_EXPLICIT_SCHED );

#if TIMER_ACTIVE
{
int timerid;
struct itimerspec value;

	event.sigev_notify		= SIGEV_PULSE;
	event.sigev_code		= NIC_TIMER_EVENT;
	event.sigev_coid		= ext->coid;
	event.sigev_priority	= NIC_PRIORITY;

	if( timer_create( CLOCK_REALTIME, &event, &timerid ) == -1 ) {
		return( -1 );
	}

	value.it_value.tv_sec	= value.it_interval.tv_sec	= 1;
	value.it_value.tv_nsec	= value.it_interval.tv_nsec	= 0;

	if( timer_settime( timerid, 0, &value, NULL ) == -1 ) {
		return( -1 );
	}
}
#endif

	event.sigev_notify		= SIGEV_PULSE;
	event.sigev_coid		= ext->coid;
	event.sigev_code		= NIC_INTERRUPT_EVENT;
	event.sigev_priority	= NIC_PRIORITY;

	if( ( ext->iid = InterruptAttachEvent( nic->cfg.IRQRegisters[0], &event, _NTO_INTR_FLAGS_TRK_MSK ) ) == -1 ) {
		fprintf( stderr, "InterruptAttachEvent failed\n" );
		perror( "InterruptAttachEvent failed\n" );
		return( 1 );
	}

	// create the interface thread
	if( pthread_create( &ext->tid, &pattr, (void *)dm9pci_driver_thread, nic ) ) {
		fprintf( stderr, "dm9pci_attach:  Unable to create driver thread\n" );
		return( 1 );
	}

	if( dm9pci_config( nic ) ) {
		return( 1 );
	}


       dm9pci_entry.func_hdl = (void *) nic;
       if (ion_register(dll_hdl, &dm9pci_entry, &ext->reg_hdl, &ext->cell,
                        &lan) < 0) {
               return (1);
	}
	nic->lan = lan;

        if (!ext->max_pkts)
                ext->max_pkts = 100;
        ext->ion->devctl(ext->reg_hdl, DCMD_IO_NET_MAX_QUEUE,
                        &ext->max_pkts, sizeof(ext->max_pkts), NULL);

	return( 0 );
}

int dm9pci_devctl( void *hdl, int dcmd, void *data, size_t size, int *ret )
{
	Nic_t				*nic = (Nic_t *)hdl;
	Nic_Dm9pci_Ext_t		*ext;
	int					status;

#if DEBUG_DETECT
fprintf(dfd, "dm9pci_devctl()\n\r");
#endif
	status	= EOK;
	ext		= (Nic_Dm9pci_Ext_t *)nic->ext;

	switch( dcmd ) {
		case DCMD_IO_NET_NICINFO:
			memcpy( data, nic, min( size, sizeof( Nic_t ) ) );
			break;

        case DCMD_IO_NET_TXRX_COUNT:
            memcpy( data, &ext->txrx_count,
                    min( size, sizeof( NicTxRxCount_t ) ) );
            break;

		default:
			status = ENOTSUP;
			break;
	}

	return( status );
}

int dm9pci_cntrs( Nic_t *nic )
{
	int				base;
	EthernetStats_t	*estats;
	Nic_Dm9pci_Ext_t	*ext;

#if DEBUG_DETECT
fprintf(dfd, "dm9pci_cntrs()\n\r");
#endif
	ext			= (Nic_Dm9pci_Ext_t *)nic->ext;
	base		= ext->iobase;
	estats		= &nic->nstats.un.estats;
	estats->rcv_overrun			+= in8( base + OVER_CNT );
	estats->rcv_alignment_error	+= in8( base + ALIGN_CNT );
	estats->rcv_crc_error		+= in8( base + CRC_ERR_CNT );

	return( 0 );
}

int dm9pci_process_interrupt( Nic_t *nic )
{
	Nic_Dm9pci_Ext_t		*ext;

#if DEBUG_INT
fprintf(dfd, "dm9pci_process_interrupt()\n\r");
#endif
	ext			= (Nic_Dm9pci_Ext_t *)nic->ext;

	/* disable all interrupts in CSR7 to solve the interrupt edge problem */	
	OUT32(ext, CSR7, 0);

	/* Get and clear status */

	ext->cr5_data = IN32(ext, CSR5);
	
	while ( (ext->cr5_data = IN32(ext, CSR5)) & CSR7_DEFAULT )
	{
		OUT32(ext, CSR5, ext->cr5_data);	/* clear status register */

		if (ext->cr5_data & CSR5_RXCI)	// Rx interrupt
			dm9pci_receive( nic );

		if (ext->cr5_data & CSR5_TXCI)	// Tx interrupt
			dm9pci_transmit_complete( nic );
	}

	/* restore CSR7 to re-enable interrupt */
	OUT32(ext, CSR7, ext->cr7_data);

	InterruptUnmask( nic->cfg.IRQRegisters[0], ext->iid );
	return( 0 );
}

int dm9pci_event_handler( Nic_t *nic )
{
	struct _pulse		pulse;
	iov_t				iov;
	int					rcvid;
	Nic_Dm9pci_Ext_t	*ext;

#if DEBUG_INT
fprintf(dfd, "dm9pci_event_handler()\n\r");
#endif
	ext = (Nic_Dm9pci_Ext_t *)nic->ext;
	SETIOV( &iov, &pulse, sizeof( pulse ) );

	while( 1 ) {
		if( ( rcvid = MsgReceivev( ext->chid, &iov, 1, NULL ) ) == -1 ) {
                        if( errno == ESRCH ) {
                                pthread_exit( NULL );
                        }
                        continue;
		}
                
		pthread_mutex_lock( &ext->mutex );
		switch( pulse.code ) {
			case NIC_INTERRUPT_EVENT:
				dm9pci_process_interrupt( nic );
				break;

#if TIMER_ACTIVE
			case NIC_TIMER_EVENT:
				dm9pci_timer( nic );
				break;
#endif

//			case MDI_TIMER:
//				MDI_MonitorPhy( ext->mdi );
//				break;

			default:
				if( rcvid ) {
					MsgReplyv( rcvid, ENOTSUP, &iov, 1 );
				}
				break;
		}
		pthread_mutex_unlock( &ext->mutex );
	}
	return( 0 );
}

void *dm9pci_driver_thread( void *data )
{
#if DEBUG_INT
fprintf(dfd, "dm9pci_driver_thread()\n\r");
#endif
	dm9pci_event_handler( (Nic_t *)data );
	return( NULL );
}

int dm9pci_shutdown1( int reg_hdl, void *hdl )
{
	Nic_t			*nic = (Nic_t *)hdl;

	Nic_Dm9pci_Ext_t	*ext;

#if DEBUG_INT
fprintf(dfd, "dm9pci_shutdown1()\n\r");
#endif
	ext = (Nic_Dm9pci_Ext_t	*)nic->ext;

	ext->open = 0;
	OUT32(ext, CSR7, 0);		/* disable interrupts */
	OUT32(ext, CSR6, CSR6_DEFAULT);	/* stop Tx/Rx */

	InterruptDetach( ext->iid );
	ConnectDetach( ext->coid );
	ChannelDestroy( ext->chid );

	return (0);
}

int dm9pci_shutdown2(int reg_hdl, void *hdl)
{
	Nic_t		*nic = (Nic_t *)hdl;
	int			cnt;
	npkt_t		*npkt;

	Nic_Dm9pci_Ext_t  *ext;

#if DEBUG_DETECT
fprintf(dfd, "dm9pci_shutdown2()\n\r");
#endif
	ext = (Nic_Dm9pci_Ext_t   *)nic->ext;

	pthread_join(ext->tid, NULL);

	// check for used rx buffers and free them
	for( cnt = 0; cnt < RX_DESC_NUM; cnt++ ) {
		if( npkt = ext->rx_pktq[cnt] )
		{
			ion_free(npkt->org_data);
			ion_free(npkt);
			ext->rx_pktq[cnt] = NULL;
		}
	}

	for( ; npkt = ext->rx_free; ) {
		ext->rx_free = npkt->next;
		ion_free(npkt->org_data);
		ion_free(npkt);
	}

	dm9pci_flush( ext->reg_hdl, nic );

	pthread_mutex_destroy( &ext->mutex );

	pci_detach_device(ext->hdl);

	munmap( (void *)ext->rfd, sizeof(dm9pci_rx_desc) * RX_DESC_NUM + 31 );
	munmap( (void *)ext->tfd, sizeof(dm9pci_tx_desc) * TX_DESC_NUM + 31 );
	munmap( (void *)ext->alloc_tx_buf, BUFFER_SIZE * TX_DESC_NUM + 31 );

	free( ext->rx_pktq );
	free( ext->tx_pktq );
	free( nic->ext );
	free( nic );

	return( 0 );
}

int dm9pci_advertise(int reg_hdl, void *func_hdl)
{
	npkt_t                                  *npkt;
	net_buf_t                               *nb;
	net_iov_t                               *iov;
	Nic_t                                   *nic;
	Nic_Dm9pci_Ext_t  *ext;
	io_net_msg_dl_advert_t                  *ap;

#if DEBUG_DETECT
fprintf(dfd, "dm9pci_advertise()\n\r");
#endif
	nic = (Nic_t *)(func_hdl);
	ext = (Nic_Dm9pci_Ext_t   *)nic->ext;

	/* Use 1520 here because may go on rx_free list when dm9pci_receive_complete called */
	if( (npkt = dm9pci_alloc_npkt(nic, BUFFER_SIZE)) == NULL ) {
#if DEBUG_DETECT
fprintf(dfd, "allocate npkt fail!!!\n\r");
#endif
		return 0;
	}

	nb = npkt->buffers.tqh_first;
	iov = nb->net_iov;

//	(int)iov->iov_base &= ~31;  //realign to dword.
	ap = iov->iov_base;
	iov->iov_len = sizeof *ap; //reset this

	memset( ap, 0x00, sizeof *ap );
	ap->type   = _IO_NET_MSG_DL_ADVERT;

    ap->iflags = (IFF_SIMPLEX | IFF_BROADCAST | IFF_RUNNING);
    if( nic->flags & NIC_FLAG_MULTICAST )
    {
        ap->iflags |= IFF_MULTICAST;
    }

	ap->mtu_min       = 0;
	ap->mtu_max       = nic->mtu;
	ap->mtu_preferred = nic->mtu;
	strcpy(ap->up_type, "en");
	itoa(nic->lan, ap->up_type + 2, 10);

	strcpy(ap->dl.sdl_data, ap->up_type);

	ap->dl.sdl_len = sizeof(struct sockaddr_dl);
	ap->dl.sdl_family = AF_LINK;
	ap->dl.sdl_index  = nic->lan;
	ap->dl.sdl_type = IFT_ETHER;
	ap->dl.sdl_nlen = strlen(ap->dl.sdl_data); /* not null terminated */
	ap->dl.sdl_alen = 6;
	memcpy(ap->dl.sdl_data + ap->dl.sdl_nlen, nic->current_address, 6);

	npkt->flags |= _NPKT_MSG;
	npkt->iface = 0;
	npkt->framelen = sizeof *ap;
#if DEBUG_DETECT
fprintf(dfd, "ivoke ion_add_done()\n\r");
#endif
	if(ion_add_done( ext->reg_hdl, npkt, nic ) == -1) {
#if DEBUG_DETECT
fprintf(dfd, "ion_add_done() fail!!!\n\r");
#endif
		ion_free(npkt->org_data);
		ion_free(npkt);
		return(0);
	}
#if DEBUG_DETECT
fprintf(dfd, "invoke ion_rx_packets()\n\r");
#endif
	if(ion_rx_packets(ext->reg_hdl, npkt, 0, 0, ext->cell, nic->lan, 0) == 0)
	{
#if DEBUG_DETECT
fprintf(dfd, "ivoke ion_tx_complete()\n\r");
#endif
		ion_tx_complete(ext->reg_hdl, npkt);
	}


	return(0);
}

int dm9pci_flush( int reg_hdl, void *hdl )
{
	Nic_t			*nic = (Nic_t *)hdl;
	Nic_Dm9pci_Ext_t	*ext;
	npkt_t			*npkt;
	npkt_t			*tmp;

#if DEBUG_DETECT
fprintf(dfd, "dm9pci_flush()\n\r");
#endif
	ext = (Nic_Dm9pci_Ext_t	*)nic->ext;

	pthread_mutex_lock( &ext->mutex );

	npkt = ext->nhead;
	ext->nhead = ext->ntail = NULL;

	for( ; npkt; npkt = tmp ) {

		tmp = npkt->next;
               	ion_tx_complete(ext->reg_hdl, npkt);
        }

	pthread_mutex_unlock( &ext->mutex );

	return( 0 );
}

npkt_t *dm9pci_alloc_npkt( Nic_t *nic, size_t size )
{
	Nic_Dm9pci_Ext_t		*ext;
	npkt_t                          *npkt;
	net_buf_t                       *nb;
	net_iov_t                       *iov;
	char                            *ptr;

#if DEBUG_DETECT
fprintf(dfd, "dm9pci_alloc_npkt()\n\r");
#endif
	ext			= (Nic_Dm9pci_Ext_t *)nic->ext;

	if((npkt = ion_alloc_npkt(sizeof(net_buf_t) + sizeof(net_iov_t),
		(void **) &nb)) == NULL) {

		return( NULL );
	}

	if((ptr = ion_alloc(size + 31, 0)) == NULL) {
		ion_free(npkt);
		return(NULL);
	}

	TAILQ_INSERT_HEAD(&npkt->buffers, nb, ptrs);

	iov = (net_iov_t *)(nb + 1);

	nb->niov                = 1;
	nb->net_iov             = iov;

	iov->iov_base	= (char *) ((uint32_t)(ptr+31) & ~31);
	iov->iov_len    = size;
	iov->iov_phys   = (paddr_t)(ion_mphys(iov->iov_base));

	npkt->org_data  = ptr;
	npkt->next      = NULL;
	npkt->tot_iov   = 1;

	return (npkt);
}

/************************************************************************
 *
 *			Descriptor Ring Functions
 *
 ************************************************************************/
int dm9pci_alloc_rx_buf(Nic_t *nic, dm9pci_rx_desc *rx_desc_ptr)
{
	npkt_t			*npkt;
	net_buf_t		*nb;
	net_iov_t		*iov;

#if DEBUG_RX
fprintf(dfd, "dm9pci_alloc_rx_buf()\n\r");
#endif
	npkt = dm9pci_alloc_npkt(nic, BUFFER_SIZE);
	npkt->next = NULL;
	rx_desc_ptr->rx_npkt_ptr = (uint32_t)npkt;

	nb = npkt->buffers.tqh_first;
	iov = nb->net_iov;

	rx_desc_ptr->rdes2 = iov->iov_phys;
	rx_desc_ptr->rx_buf_ptr = (uint32_t)iov->iov_base;
	
	return 0;
}

int dm9pci_descriptor_init(Nic_t *nic)
{
	int cnt;
	Nic_Dm9pci_Ext_t *ext;
	dm9pci_rx_desc *rfd;
	dm9pci_tx_desc *tfd;
	net_buf_t *buf;
	net_iov_t *iov;
	npkt_t *npkt;
	uint32_t tfdp=0, rfdp=0;

#if DEBUG_INIT
fprintf(dfd, "dm9pci_descriptor_init()\n\r");
#endif
	ext = (Nic_Dm9pci_Ext_t *)nic->ext;

	/*
	 * build circular linked list of Rx descriptors
	 */
	ext->rx_desc = (dm9pci_rx_desc *)((uint32_t)(ext->rfd+31) & ~31);
	for (cnt=0, rfd=ext->rx_desc; cnt < RX_DESC_NUM; cnt++, rfd++)
	{
		ext->rx_pktq[cnt] = dm9pci_alloc_npkt(nic, BUFFER_SIZE);
		
		npkt = ext->rx_pktq[cnt];
		buf = TAILQ_FIRST(&npkt->buffers);
		iov = buf->net_iov;

		rfd->rdes0 = RDES0_OWN;		/* owned by MAC */	
		rfd->rdes1 = RDES1_CE | MAX_FRAME_SIZE;	/* chain enable */
		rfd->rdes2 = iov->iov_phys;
		if (cnt == (RX_DESC_NUM-1))
		{
			rfdp = (uint32_t)mphys((void *)&ext->rx_desc[0]);
			rfd->rdes3 = rfdp;
		}
		else
			rfd->rdes3 = mphys((void *)&ext->rx_desc[cnt+1]);
	}

	/* set Receive Descriptor Base Address */
	OUT32(ext, CSR3, rfdp);

	/* reset Rx producer & consumer index */
	ext->rx_pidx = ext->rx_cidx = 0;
#if DEBUG_RX
dump_rx_desc(ext);
#endif

	/*
	 * build circular linked list of Tx descriptors
	 */
	ext->tx_desc = (dm9pci_tx_desc *)((uint32_t)(ext->tfd+31) & ~31);
	ext->tx_buf = (unsigned char *)((uint32_t)(ext->alloc_tx_buf+31) & ~31);
	
	for (cnt=0, tfd=ext->tx_desc; cnt < TX_DESC_NUM; cnt++, tfd++)
	{
		tfd->tdes0 = 0;		/* owned by host */
		tfd->tdes1 = TDES1_CI | TDES1_CE;	/* completion int & chain enable */
		tfd->tdes2 = mphys(&ext->tx_buf[cnt * BUFFER_SIZE]);
		if (cnt == (TX_DESC_NUM-1))
		{
			tfdp = (uint32_t)mphys((void *)&ext->tx_desc[0]);
			tfd->tdes3 = tfdp;
		}
		else
			tfd->tdes3 = mphys((void *)&ext->tx_desc[cnt+1]);
	}
	
	/* set Transmit Descriptor Base Address */
	OUT32(ext, CSR4, tfdp);

	/* reset Tx producer & consumer index */
	ext->tx_pidx = ext->tx_cidx = (TX_DESC_NUM -1);
#if DEBUG_TX
dump_tx_desc(ext);
#endif

	return(0);	
}

void dm9pci_descriptor_free(Nic_Dm9pci_Ext_t *ext)
{
//fprintf(dfd, "dm9pci_descriptor_free()\n\r");
//	delete_area(ext->tx_buf_area);
//	delete_area(ext->rx_buf_area);
}

#if CONSOLE_DEBUG
void dump_tx_desc(Nic_Dm9pci_Ext_t *ext)
{
	int	i;

	fprintf(dfd, "CSR5 = %08X", IN32(ext, CSR5));
	fprintf(dfd, ", CSR6 = %08X", IN32(ext, CSR6));
	fprintf(dfd, ", CSR7 = %08X\n\r", IN32(ext, CSR7));
	fprintf(dfd, "CSR4 = %08X", IN32(ext, CSR4));
	fprintf(dfd, ", cur_tx = %d\n\r", ext->cur_tx);

	for(i=0; i< TX_DESC_NUM; i++)
		fprintf(dfd, "tx_desc[%d] = %08x %08x %08x %08x\n\r", i,
			ext->tx_desc[i].tdes0,
			ext->tx_desc[i].tdes1,
			ext->tx_desc[i].tdes2,
			ext->tx_desc[i].tdes3);
}

void dump_rx_desc(Nic_Dm9pci_Ext_t *ext)
{
	int	i;

	fprintf(dfd, "CSR3 = %08x, rx_pidx = %d, rx_cidx = %d\n\r",
			IN32(ext, CSR3), ext->rx_pidx, ext->rx_cidx);
	for(i=0; i< RX_DESC_NUM; i++)
		fprintf(dfd, "rx_desc[%d] = %08x %08x %08x %08x\n\r", i,
				ext->rx_desc[i].rdes0,
				ext->rx_desc[i].rdes1,
				ext->rx_desc[i].rdes2,
				ext->rx_desc[i].rdes3);

}
#endif

static void init_HPNA_PHY(Nic_Dm9pci_Ext_t *ext)
{
	uint16_t pna_cr16, tmp;

#if DEBUG_AUTO
fprintf(dfd, "init_HPNA_PHY()\n\r");
#endif
	/* CSR6[18]=1, select external PHY */
	update_cr6(ext,  (ext->cr6_data | CSR6_PS));
	
	if( (phy_read(ext, ext->phy_addr, PHY_PHYIDR2) & 0xFFF0) == 0xB900 )
	{
		/* CR16 default = ignore remote command/high speed/low power */
		pna_cr16 = 0x8005;
		
		/* PNA present, check if it's HOMERUN or LONGRUN */
		if( phy_read(ext, ext->phy_addr, 31) == 0x110F )
		{
//fprintf(dfd, "There's a LongRun PNA!\n\r");
			ext->HPNA_type = HPNA_LONGRUN;
			
			phy_write(ext, ext->phy_addr, PHY_BMCR, BMCR_RST);
			/* wait for 4 seconds until PNA PHY stable */	
			sleep(4);
			
			/* enable %25 increase of noise slice for LONGRUN */
			pna_cr16 &= ~0x1000;
			phy_write(ext, ext->phy_addr, 16, pna_cr16);
			
			/* set low byte of Reg. 25 to 02h */
			tmp = phy_read(ext, ext->phy_addr, 25);
			tmp &= 0xFF00;
			phy_write(ext, ext->phy_addr, 25, tmp | 0x0002);
		}
		else
		{
//fprintf(dfd, "There's a HomeRun PNA!\n\r");
			ext->HPNA_type = HPNA_HOMERUN;
			
			/* disable %25 increase of noise slice for HOMERUN */
			pna_cr16 |= 0x1000;
			phy_write(ext, ext->phy_addr, 16, pna_cr16);

			/* noise ceiling(CR25[8..15]) = 0xF0,
			   noise floor(CR25[0..7]) = CR24[0..7] + 3 */
			tmp = phy_read(ext, ext->phy_addr, 24);
			tmp += 3;
			tmp &= 0x00FF;
			tmp |= 0xF000;
			phy_write(ext, ext->phy_addr, 25, tmp);
		}
		
	}
	else
{
//fprintf(dfd, "Home PNA not present!\n\r");
		ext->HPNA_type = HPNA_NOT_PRESENT;
}

	/* restore CSR6 */
	update_cr6(ext, ext->cr6_data);
}

static void init_DM9102AF_E3(Nic_Dm9pci_Ext_t *ext)
{
	if ((ext->pci_pmr & 0x00070000) == 0x00010000)
	{
		/* CSR6[18]=0, select 10/100M */
		update_cr6(ext, ext->cr6_data & ~CSR6_PS);

		/* it's DM9102A E3, must power off/on PHY to simulate plug out/in */
		phy_write(ext, ext->phy_addr, PHY_BMCR, 0x1800);	/* power down PHY */
		sleep(2);		/* last for about 2 seconds */
		phy_write(ext, ext->phy_addr, PHY_BMCR, 0x1000);	/* power on PHY again */		
	
		/* restore CSR6 */
		update_cr6(ext, ext->cr6_data);
	}
}

/*
 * local variables init
 */
static void local_var_init(Nic_Dm9pci_Ext_t *ext)
{
#if DEBUG_INIT
fprintf(dfd, "local_var_init()\n\r");
#endif
	ext->open = 0;
	
	ext->cr0_data = CSR0_DEFAULT;
	ext->cr6_data = CSR6_DEFAULT;
	ext->cr7_data = CSR7_DEFAULT;
	ext->cr15_data = CSR15_DEFAULT;

	ext->num_rx_free = 0;
}

static void set_PHY_ANAR(Nic_Dm9pci_Ext_t *ext)
{
	if( ext->cfg_mode == MEDIA_TYPE_AUTO )
	{
		if( (ext->chip_id == PCI_DEVICE_ID_DM9009) ||
		   ( (ext->srom[18] == 0x14) &&
		     ( (*(uint32_t *)&ext->srom[10] == 0x90091282) &&
		       (*(uint32_t *)&ext->srom[0] == 0x82120990) ) ) )
		{
#if DEBUG_AUTO
fprintf(dfd, "IT'S A DM9009!!!!!\n\r");
#endif
			/* DM9009 support only 10M full/half duplex */
			ext->anar = (ANAR_FCS | ANAR_10_FDX | ANAR_10_HDX | ANAR_CSMACD);
		}
		else
			/* default support all capability */
			ext->anar = (ANAR_FCS | ANAR_CAP_ALL | ANAR_CSMACD);
	}
	else
	{
		switch(ext->cfg_mode)
		{
		case MEDIA_TYPE_100FDX:
			ext->anar = (ANAR_FCS | ANAR_TX_FDX | ANAR_CSMACD);
			break;
		case MEDIA_TYPE_100HDX:
			ext->anar = (ANAR_FCS | ANAR_TX_HDX | ANAR_CSMACD);
			break;
		case MEDIA_TYPE_10FDX:
			ext->anar = (ANAR_FCS | ANAR_10_FDX | ANAR_CSMACD);
			break;
		case MEDIA_TYPE_10HDX:
			ext->anar = (ANAR_FCS | ANAR_10_HDX | ANAR_CSMACD);
			break;
		default:
			ext->anar = (ANAR_FCS | ANAR_CAP_ALL | ANAR_CSMACD);
			break;
		}
	}

	/* CSR6[18]=0, select 10/100M */
	update_cr6(ext, ext->cr6_data & ~CSR6_PS);

	phy_write(ext, ext->phy_addr, PHY_ANAR, ext->anar);

/* restart auto negotiation */
phy_write(ext, ext->phy_addr, PHY_BMCR,
		phy_read(ext, ext->phy_addr, PHY_BMCR) | BMCR_RAN);

	/* restore CSR6 */
	update_cr6(ext, ext->cr6_data);
}

static void startup_auto_detect(Nic_t *nic)
{
	Nic_Dm9pci_Ext_t *ext = (Nic_Dm9pci_Ext_t *)nic->ext;

	ext->phy_addr = 1;

	/* reset PHY through General Purpose Register (CSR12) */
	OUT32(ext, CSR12, CSR12_GEPC|CSR12_GEPD);	/* let bit 7 output port */
	OUT32(ext, CSR12, CSR12_GEPD);		/* raise reset PHY signal */
	OUT32(ext, CSR12, 0);			/* clear reset signal */

	set_PHY_ANAR(ext);

	/* check if HOME-PNA PHY is there and setup it */ 
	init_HPNA_PHY(ext);
	
	/* DM9102AF E3 will not stop its autonegotiation with some PHY (NS83840).
	   It will be necessary to power down PHY for 2 seconds and then power on
	   it again to simulate plug UTP line out/in */
	init_DM9102AF_E3(ext);

	/* startup auto detect media type */
	auto_detect_media_type(nic);
}

/*
 * Auto detect media mode
 */
static void auto_detect_media_type(Nic_t *nic)
{
	int i;
	uint16_t phy_mode = 0;
	int speed = 0;
	Nic_Dm9pci_Ext_t *ext = (Nic_Dm9pci_Ext_t *)nic->ext;

#if DEUBG_AUTO
fprintf(dfd, "auto_detect_media_type()\n\r");
#endif

	/* CSR6[18]=0, select 10/100M */
	update_cr6(ext, ext->cr6_data & ~CSR6_PS);

	for(i=0; i<5; i++)
	{
		phy_mode = phy_read(ext, ext->phy_addr, PHY_BMSR);
		if ( (phy_mode & (BMSR_ANC|BMSR_LS)) == (BMSR_ANC|BMSR_LS) )
			break;
		sleep(1);
	}

#if DEBUG_AUTO
#endif
	if ( (phy_mode & (BMSR_ANC|BMSR_LS)) == (BMSR_ANC|BMSR_LS) )
	{
#if DEBUG_AUTO
fprintf(dfd, "link pass\n\r");
#endif
		ext->link_status = LINK_PASS;

		phy_mode = phy_read(ext, ext->phy_addr, PHY_DSCR) & DSCR_M_OM;
		
		switch (phy_mode)
		{
		case 0x1000: ext->op_mode=MEDIA_TYPE_10HDX; speed=10; break;
		case 0x2000: ext->op_mode=MEDIA_TYPE_10FDX; speed=10; break;
		case 0x4000: ext->op_mode=MEDIA_TYPE_100HDX; speed=100; break;
		case 0x8000: ext->op_mode=MEDIA_TYPE_100FDX; speed=100; break;
		default: ext->op_mode=MEDIA_TYPE_10HDX; speed=10;
                break;
		}
	}
	else
	{
#if DEBUG_AUTO
fprintf(dfd, "link fail\n\r");
#endif
		ext->link_status = LINK_FAIL;

		if (ext->HPNA_type == HPNA_NOT_PRESENT) {
			ext->op_mode=MEDIA_TYPE_10HDX;
			speed = 0;
		}	
		else {
			ext->op_mode=MEDIA_TYPE_1HOMEPNA;
			speed = 1;
		}
	}
	
	nic->media_rate = speed * 1000;

#if DEBUG_AUTO
fprintf(dfd, "Media Type = ");
switch(ext->op_mode)
{
case MEDIA_TYPE_10HDX: fprintf(dfd, "10HDX"); break;
case MEDIA_TYPE_10FDX: fprintf(dfd, "10FDX"); break;
case MEDIA_TYPE_100HDX: fprintf(dfd, "100HDX"); break;
case MEDIA_TYPE_100FDX: fprintf(dfd, "100FDX"); break;
case MEDIA_TYPE_1HOMEPNA: fprintf(dfd, "1HOMEPNA"); break;
}
fprintf(dfd, "\n\r");
#endif
	set_CSR6_op_mode(ext);
}

/*
 * Process op-mode
 * AUTO mode : PHY controller in Auto-negotiation Mode
 * Force mode: PHY controller in force mode with HUB
 *             N-way force capability with SWITCH
 */
static void set_CSR6_op_mode(Nic_Dm9pci_Ext_t *ext)
{
#if DEBUG_AUTO
//fprintf(dfd, "set_CSR6_op_mode()\n\r");
#endif
	/* Full Duplex Mode Check */
	if (ext->op_mode == MEDIA_TYPE_100FDX ||
	    ext->op_mode == MEDIA_TYPE_10FDX)
		ext->cr6_data |= CSR6_FDM;	/* Set Full Duplex Bit */
	else
		ext->cr6_data &= ~CSR6_FDM;	/* Clear Full Duplex Bit */

	/* Transciver Selection */
	if (ext->op_mode == MEDIA_TYPE_1HOMEPNA) /* 1M HomeRun or 1M LongRun */
		ext->cr6_data |= CSR6_PS;	/* External HomeRun or LongRun */
	else
		ext->cr6_data &= ~CSR6_PS;	/* Internal 10/100 transciver */
 
	update_cr6(ext, ext->cr6_data);
}


/*
 * Update both CSR6 and its local mirror vaule
 * Firstly stop MAC Tx/Rx, update value and then restore Tx/Rx
 */
static void update_cr6(Nic_Dm9pci_Ext_t *ext, uint32_t cr6_data)
{
	uint32_t cr6_tmp;

#if DEBUG_AUTO
fprintf(dfd, "update_cr6()\n\r");
#endif
	cr6_tmp = cr6_data & ~(CSR6_TXSC|CSR6_RXSC);	/* stop Tx/Rx */
	OUT32(ext, CSR6, cr6_tmp);
	DELAY_5US;
	OUT32(ext, CSR6, cr6_data);
	DELAY_5US;
	//cr6_tmp=IN32(ext, CSR6);
	//dprintf("CSR6 update %x\n\r", cr6_tmp);
}

/*
 * Send a setup frame for DM9102/DM9102A
 * This 192 bytes long setup frame initilize DM910X addres filter mode
 */
static int send_setup_frame(Nic_Dm9pci_Ext_t *ext)
{
	int		pidx;
	uint16_t *eaddrs;
	uint32_t	*setup_frm;
	int i;

#if DEBUG_INIT
fprintf(dfd, "send_setup_frame()\n\r");
#endif
	eaddrs = (uint16_t *)ext->macaddr.ebyte;

	/* use current Tx descriptor to send setup frame */
	pidx = (ext->tx_pidx + 1) % TX_DESC_NUM;
	setup_frm = (uint32_t *)&ext->tx_buf[pidx];

	/* Fill our MAC address for perfect filtering */
	*setup_frm++ = eaddrs[0];
	*setup_frm++ = eaddrs[1];
	*setup_frm++ = eaddrs[2];

	/* Fill the rest with broadcast address to accept broadcast packet */
	for (i = 1; i < 16; i++) {
		*setup_frm++ = 0xFFFF;
		*setup_frm++ = 0xFFFF;
		*setup_frm++ = 0xFFFF;
	}

	/* prepare the setup frame Tx descriptor */
	ext->tx_desc[pidx].tdes1 = TDES1_CI | TDES1_SETF | TDES1_CE | 192;
	ext->tx_desc[pidx].tdes0 = TDES0_OWN;

	ext->tx_pidx = pidx;
	ext->tx_pktq[pidx] = NULL;
	ext->tx_cidx = ext->tx_pidx;	/* setup frame has no npkt to be released */

	/* Start the chip's Tx to process setup frame. */
	OUT32(ext, CSR6, ext->cr6_data | CSR6_TXSC);
	OUT32(ext, CSR1, 0x1);	/* Kick Tx poll demand */
//fprintf(dfd, "CSR6 = %08X after kill tx poll demnad\n\r", IN32(ext, CSR6));

	nsec_delay(500000);

#if DEBUG_TX
dump_tx_desc(ext);
#endif
	if(ext->tx_desc[pidx].tdes0 & TDES0_OWN)
		return 1;
	else
		return 0;
}

/************************************************************************
 *
 *			MII PHY Functions
 *
 ************************************************************************/
/*
 * Write a word to Phy register
 */
static void phy_write(Nic_Dm9pci_Ext_t *ext, uint8_t phy_addr, uint8_t offset, uint16_t phy_data)
{
	uint16_t i;

	/* DM9102/DM9102A Chip */
	/* Send 33 synchronization clock to Phy controller */
	for (i=0; i<35; i++)
		phy_write_1bit(ext, PHY_DATA_1);

	/* Send start command(01) to Phy */
	phy_write_1bit(ext, PHY_DATA_0);
	phy_write_1bit(ext, PHY_DATA_1);

	/* Send write command(01) to Phy */
	phy_write_1bit(ext, PHY_DATA_0);
	phy_write_1bit(ext, PHY_DATA_1);

	/* Send Phy addres */
	for (i=0x10; i>0; i=i>>1)
		phy_write_1bit(ext, phy_addr&i ? PHY_DATA_1: PHY_DATA_0);

	/* Send register addres */
	for (i=0x10; i>0; i=i>>1)
		phy_write_1bit(ext, offset&i ? PHY_DATA_1: PHY_DATA_0);

	/* written trasnition */
	phy_write_1bit(ext, PHY_DATA_1);
	phy_write_1bit(ext, PHY_DATA_0);

	/* Write a word data to PHY controller */
	for (i=0x8000; i>0; i>>=1)
		phy_write_1bit(ext, phy_data&i ? PHY_DATA_1: PHY_DATA_0);
}

/*
 * Read a word data from phy register
 */
static uint16_t phy_read(Nic_Dm9pci_Ext_t *ext, uint8_t phy_addr, uint8_t offset)
{
	int i;
	uint16_t phy_data;

	/* DM9102/DM9102A Chip */
	/* Send 33 synchronization clock to Phy controller */
	for (i=0; i<35; i++)
		phy_write_1bit(ext, PHY_DATA_1);

	/* Send start command(01) to Phy */
	phy_write_1bit(ext, PHY_DATA_0);
	phy_write_1bit(ext, PHY_DATA_1);

	/* Send read command(10) to Phy */
	phy_write_1bit(ext, PHY_DATA_1);
	phy_write_1bit(ext, PHY_DATA_0);

	/* Send Phy addres */
	for (i=0x10; i>0; i=i>>1)
		phy_write_1bit(ext, phy_addr&i ? PHY_DATA_1: PHY_DATA_0);

	/* Send register addres */
	for (i=0x10; i>0; i=i>>1)
		phy_write_1bit(ext, offset&i ? PHY_DATA_1: PHY_DATA_0);

	/* Skip transition state */
	phy_read_1bit(ext);

	/* read 16bit data */
	for (phy_data=0, i=0; i<16; i++)
	{
		phy_data<<=1;
		phy_data|=phy_read_1bit(ext);
	}

	return phy_data;
}

/*
 * Write one bit data to Phy Controller
 */
static void phy_write_1bit(Nic_Dm9pci_Ext_t *ext, uint32_t phy_data)
{
	OUT32(ext, CSR9, phy_data);		/* MII Clock Low */
	DELAY_1US;
	OUT32(ext, CSR9, phy_data|MDCLKH);	/* MII Clock High */
	DELAY_1US;
	OUT32(ext, CSR9, phy_data);		/* MII Clock Low */
	DELAY_1US;
}

/*
 * Read one bit phy data from PHY controller
 */
static uint16_t phy_read_1bit(Nic_Dm9pci_Ext_t *ext)
{
	uint16_t phy_data;

	OUT32(ext, CSR9, 0x50000);
	DELAY_1US;
	phy_data = (IN32(ext, CSR9)>>19) & 0x1;
	OUT32(ext, CSR9, 0x40000);
	DELAY_1US;

	return phy_data;
}


/************************************************************************
 *
 *			SROM Functions
 *
 ************************************************************************/
void SROM_CLK_WRITE(Nic_Dm9pci_Ext_t *ext, int offset, uint8_t data)
{
	OUT32(ext, offset, data|CSR9_SROM_READ|CSR9_SRCS);
	DELAY_5US;
	OUT32(ext, offset, data|CSR9_SROM_READ|CSR9_SRCS|CSR9_SRCLK);
	DELAY_5US;
	OUT32(ext, offset, data|CSR9_SROM_READ|CSR9_SRCS);
	DELAY_5US;
}

/*
 * Read one word data from the serial ROM
 */
static uint16_t read_srom_word(Nic_Dm9pci_Ext_t *ext, int offset)
{
	int i;
	uint16_t srom_data = 0;

	OUT32(ext, CSR9, CSR9_SROM_READ);
	OUT32(ext, CSR9, CSR9_SROM_READ | CSR9_SRCS);

	/* Send the Read Command 110b */
	SROM_CLK_WRITE(ext, CSR9, SROM_DATA_1);
	SROM_CLK_WRITE(ext, CSR9, SROM_DATA_1);
	SROM_CLK_WRITE(ext, CSR9, SROM_DATA_0);

	/* Send the offset */
	for (i = 5; i >= 0; i--)
	{
		srom_data = (offset & (1 << i)) ? SROM_DATA_1 : SROM_DATA_0;
		SROM_CLK_WRITE(ext, CSR9, srom_data);
	}

	OUT32(ext, CSR9, CSR9_SROM_READ | CSR9_SRCS);

	for (i = 16; i > 0; i--)
	{
		OUT32(ext, CSR9, CSR9_SROM_READ | CSR9_SRCS | CSR9_SRCLK);
		DELAY_5US;
		srom_data = (srom_data << 1) | ((IN32(ext, CSR9) & CSR9_CRDOUT) ? 1 : 0);
		OUT32(ext, CSR9, CSR9_SROM_READ | CSR9_SRCS);
		DELAY_5US;
	}

	OUT32(ext, CSR9, CSR9_SROM_READ);

	return srom_data;
}

#if TIMER_ACTIVE
/************************************************************************
 *
 *			Timer Handler Routine
 *
 ************************************************************************/
int dm9pci_timer( Nic_t *nic )
{
	Nic_Dm9pci_Ext_t *ext;

#if DEBUG_TIMER
fprintf(dfd, "dm9pci_timer()\n\r");
#endif
	ext = (Nic_Dm9pci_Ext_t *)nic->ext;

	if (ext->open != 1)
		return(0);

	dynamic_auto_detect(nic);

	return(0);
}

void dynamic_auto_detect(Nic_t *nic)
{
	Nic_Dm9pci_Ext_t *ext;
	uint32_t	csr12;

#if DEBUG_AUTO
//fprintf(dfd, "dynamic_auto_detect()\n\r");
#endif
	ext = (Nic_Dm9pci_Ext_t *)nic->ext;

	csr12 = IN32(ext, CSR12);
	if ( (csr12 & 0x0003) && (ext->link_status == LINK_FAIL) )
	{
		/* Link status change from FAIL to PASS */
#if DEBUG_AUTO
fprintf(dfd, "link status change from FAIL to PASS\n\r");
#endif
		ext->link_status = LINK_PASS;

		auto_detect_media_type(nic);
	}
	else if ( !(csr12 & 0x0003) && (ext->link_status == LINK_PASS) )
	{
#if DEBUG_AUTO
fprintf(dfd, "link status change from PASS to FAIL\n\r");
#endif
		/* Link status change from PASS to FAIL */
		ext->link_status = LINK_FAIL;

		if (ext->HPNA_type == HPNA_NOT_PRESENT)
			ext->op_mode=MEDIA_TYPE_10HDX;
		else
			ext->op_mode=MEDIA_TYPE_1HOMEPNA;

		set_CSR6_op_mode(ext);
	}
}
#endif
