/*
 * Copyright  2000, QNX Software Systems Ltd.  All Rights Reserved
 *
 * This source code has been published by QNX Software Systems Ltd.
 * (QSSL).  However, any use, reproduction, modification, distribution
 * or transfer of this software, or any software which includes or is
 * based upon any of this code, is only permitted under the terms of
 * the QNX Open Community License version 1.0 (see licensing.qnx.com for
 * details) or as otherwise expressly authorized by a written license 
 * agreement from QSSL.  For more information, please email 
 * licensing@@qnx.com.
 *
 */


#include <time.h>
#include <drvr/support.h>
#include <sys/slog.h>
#include <sys/slogcodes.h>

#define hexval( x ) ( isalpha( x ) ? 10 + toupper( x ) - 'A' : ( x ) - '0' )

int nic_strtomac( const char *s, unsigned char *mac )
{
	int		i;
	
	for( i = 0; i < 6; i++ ) {
		unsigned char val= 0;
		while( isspace( *s ) ) ++s;
		if( !*s || !*( s + 1 ) || !isxdigit( *s ) || !isxdigit( *( s + 1 ) ) ) {
			return( 1 );
		}
		val= 0x10 * hexval( *s ) + hexval( *( s + 1 ) );
		s += 2;
		mac[i]= val;
	}
	return( 0 );
}

int nic_calc_crc_le( char *buf, int len )
{
	ulong_t	crc;
	ulong_t	carry;
	uchar_t	data;
	uint_t	i;
	uint_t	j;

	crc = 0xffffffff;
	for( i = 0; i < len; i++ ) 
    {
		data = buf[i];
		for( j = 0; j < 8; j++ ) 
        {
            carry = ( ( data ^ crc ) & 0x01 );
			crc >>= 1;
			data >>= 1;
			if( carry ) 
            {
				crc = crc ^ 0xedb88320UL;
			}
		}
	}
	return( crc );
}


int nic_calc_crc_be( char *buf, int len )
{
	ulong_t	crc;
	ulong_t	carry;
	uchar_t	data;
	uint_t	i;
	uint_t	j;

    crc = 0xffffffff;
    for( i = 0; i < len; i++ ) 
    {
        data = buf[i];
        for( j = 0; j < 8; j++ ) 
        {
            carry = ( ( crc & 0x80000000 ) ? 1 : 0 ) ^ ( data & 0x01 );
            crc <<= 1;
            data >>= 1;
            if( carry ) 
            {
                crc = ( crc ^ 0x04c11db6 ) | carry;
            }
        }
    }
	return( crc );
}


Nic_t *nic_create_dev( int ext_size )
{
	Nic_t *nic;

	// allocate nic
	if( ( nic = (Nic_t *)calloc( 1, sizeof( Nic_t ) ) ) == NULL ) {
		slogf(_SLOG_SETCODE(_SLOGC_NETWORK,0), _SLOG_ERROR,
			"Error allocating nic device\n" );
		errno = ENOMEM;
		return( NULL );
	}

	// allocate nic driver extension
	if( ( nic->ext = (Nic_Ext_t *)calloc( 1, ext_size ) ) == NULL ) {
		slogf(_SLOG_SETCODE(_SLOGC_NETWORK,0), _SLOG_ERROR,
			"Error allocating nic device extension\n" );
		errno = ENOMEM;
		free( nic );
		return( NULL );
	}

	return( nic );
}

int nic_display_config( Nic_t *nic )
{
	int				cnt;
	Config_Info_t	*cfg;

	cfg = &nic->cfg;

	fprintf( stderr, "%s\n", cfg->Description );
	fprintf( stderr, "Vendor .............. 0x%lx\n", cfg->Device_ID.DevID & 0xffff );
	fprintf( stderr, "Device .............. 0x%lx\n", cfg->Device_ID.DevID >> 16 );
	fprintf( stderr, "Revision ............ 0x%lx\n", cfg->Device_ID.SerialNum );
	for( cnt = 0; cnt < cfg->NumIOPorts; cnt++ ) {
		fprintf( stderr, "I/O port base ....... 0x%lx\n", cfg->IOPort_Base[cnt] );
	}
	for( cnt = 0; cnt < cfg->NumMemWindows; cnt++ ) {
		fprintf( stderr, "Memory base ......... 0x%lx\n", cfg->MemBase[cnt] );
	}
	for( cnt = 0; cnt < cfg->NumIRQs; cnt++ ) {
		fprintf( stderr, "Interrupt ........... 0x%lx\n", cfg->IRQRegisters[cnt] );
	}
	for( cnt = 0; cnt < cfg->NumDMAs; cnt++ ) {
		fprintf( stderr, "DMA ................. 0x%lx\n", cfg->DMALst[cnt] );
	}
	fprintf( stderr, "MAC address ......... %04x%02x %02x%04x\n",
			(ushort_t)((nic->current_address[0] << 8) | nic->current_address[1]),
			nic->current_address[2], nic->current_address[3],
			(ushort_t)((nic->current_address[4] << 8) | nic->current_address[5]) );

	return( 0 );
}

int nic_drvr_options( Nic_t *nic, char *options )
{
	int					opt;
	int					cnt;
	char				*value;
	static char			*nic_drvr_opts[] = {
							"ioport", "irq",
							"smem", "smemsz",
							"dma", "vid",
							"did", "pci",
							"mac", "lan",
							"mtu", "duplex",
							"speed", "media",
							"nomulticast", "promiscuous",
							"connector",
							NULL
						};

	if( ( opt = getsubopt( &options, nic_drvr_opts, &value ) ) == -1 ) {
		return( EINVAL );
	}

	if( nic == (Nic_t *)NULL )
	{
		return( EINVAL );
	}

    if( value == NULL )
        value = "";

	switch( opt ) {
		case 0:
			nic->cfg.IOPort_Base[nic->cfg.NumIOPorts] = strtoul( value, 0, 0 );
			nic->cfg.NumIOPorts++;
			break;

		case 1:
			nic->cfg.IRQRegisters[nic->cfg.NumIRQs] = strtoul( value, 0, 0 );
			nic->cfg.NumIRQs++;
			break;

		case 2:
			nic->cfg.MemBase[nic->cfg.NumMemWindows] = strtoul( value, 0, 0 );
			nic->cfg.NumMemWindows++;
			break;

		case 3:
			for( cnt = 0; cnt < MAX_MEM_REGISTERS; cnt++ ) {
				if( !nic->cfg.MemLength[cnt] ) {
					break;
				}
			}
			nic->cfg.MemLength[cnt] = strtoul( value, 0, 0 );
			break;

		case 4:
			nic->cfg.DMALst[nic->cfg.NumDMAs]	= strtoul( value, 0, 0 );
			nic->cfg.NumDMAs++;
			break;

		case 5:		// vender id
			/* Clear any existing VendorId & fill it in */
			nic->cfg.Device_ID.DevID &= (0xFFFF << 16);
			nic->cfg.Device_ID.DevID |= strtoul(value, 0, 0);
			break;

		case 6:		// device id
			/* Clear any existing DendorId & fill it in */
			nic->cfg.Device_ID.DevID &= 0xFFFF;
			nic->cfg.Device_ID.DevID |= (strtoul(value, 0, 0)<< 16);
			break;

		case 7:		// pci index
			nic->cfg.Device_ID.SerialNum |= strtoul( value, 0, 0 );
			break;

		case 8:
			if( nic_strtomac( value, nic->current_address ) ) {
				fprintf( stderr, "%s is an invalid mac address\n", value );
			}
			break;

		case 9:
			nic->lan = strtoul( value, 0, 0 );
			break;

		case 10:
			nic->mtu = strtoul( value, 0, 0 );
			break;

		case 11:
			if( strtoul( value, 0, 0 ) ) {
				nic->flags |= NIC_FLAG_FDX;
			}
			break;

		case 12:
			nic->media_rate = strtoul( value, 0, 0 );
            nic->media_rate *= 1000;
			break;

		case 13:
			nic->media = strtoul( value, 0, 0 );
			break;

		case 14:
			nic->flags &= ~NIC_FLAG_MULTICAST;
			break;

		case 15:
			nic->flags |= NIC_FLAG_PROMISCUOUS;
			break;

		case 16:
			switch( strtoul( value, 0, 0 ) ) {
				case 0:
					nic->flags |= NIC_FLAG_BNC;
					break;
				case 1:
					nic->flags |= NIC_FLAG_UTP;
					break;
				case 2:
					nic->flags |= NIC_FLAG_AUI;
					break;
				case 3:
					nic->flags |= NIC_FLAG_FIBRE;
					break;
			}
			break;
	}

	return( EOK );
}

#define NS_BASE			0x00
#define NE2000_DATA		0x10		// REP IO data port
#define NE2000_RESET	0x1f		// read to reset
#define	NS_CR			( NS_BASE + 0x0 )
#define	NS_0W_IMR  	  	( NS_BASE + 0xF )	// interrupt mask reg

// command register
#define	NS_CR_STP		0x01				// software reset cmd
#define	NS_CR_STA		0x02				// start NIC
#define	NS_CR_TXP		0x04				// tx pkt
#define	NS_CR_RD_NA		0x00				// rem dma cmd: not allowed
#define	NS_CR_RD_RR		0x08				// rem dma cmd: remote read
#define	NS_CR_RD_RW		0x10				// rem dma cmd: remote write
#define	NS_CR_RD_SP		0x18				// rem dma cmd: send pkt
#define	NS_CR_RD_AB		0x20				// rem dma cmd: abort
#define	NS_CR_PS0  		0x00				// select page zero regs
#define	NS_CR_PS1  		0x40				// select page one regs
#define	NS_CR_PS2  		0x80				// select page two regs

int nic_ne2000( int iobase )
{
	uchar_t tmp;

	tmp = in8( iobase + NE2000_RESET );			// reset
	delay( 10 );
	out8( iobase + NE2000_RESET, tmp );
	delay( 10 );
	tmp = in8( iobase + NS_CR );				// read the command register
	if( ( tmp & NS_CR_STP ) != NS_CR_STP ) {	// STP powers up high
		return( 1 );
	}

	if( ( tmp & NS_CR_STA ) != 0x00 ) {			// STA powers up lo
		return( 1 );
	}
	
	out8( iobase + NS_CR, 0xA1 );				// switch to page 2
	tmp = in8( iobase + NS_0W_IMR );			// read IMR
	if( ( tmp & 0x7f) != 0 ) {					// IMR powers up zeroes
		return( 1 );
	}

	return( 0 );
}
