#include <dm9pci.h>

extern FILE *dfd;
extern void dump_npkt(npkt_t *npkt);
void dump_rx_desc(Nic_Dm9pci_Ext_t *ext);

static int dm9pci_add_rfd( Nic_t *nic, npkt_t *npkt )
{
	Nic_Dm9pci_Ext_t *ext;
	int				pidx;
	int				base;
	dm9pci_rx_desc *rfd;
	net_buf_t *buf;
	net_iov_t *iov;

#if DEBUG_RX
fprintf(dfd, "dm9pci_add_rfd()\n\r");	
#endif
	ext		= (Nic_Dm9pci_Ext_t *)nic->ext;

	base	= ext->iobase;

	pidx	= ext->rx_pidx;
	rfd		= (dm9pci_rx_desc *)&ext->rx_desc[pidx];

	ext->rx_pktq[pidx] = npkt;

	buf = TAILQ_FIRST(&npkt->buffers);
	iov = buf->net_iov;

	rfd->rdes1 = RDES1_CE | MAX_FRAME_SIZE;	/* chain enable */
	rfd->rdes2 = iov->iov_phys;
	rfd->rdes0 = RDES0_OWN;		/* owned by MAC */	

	ext->rx_pidx	= ( pidx + 1 ) % RX_DESC_NUM;

	return( 0 );
}

int dm9pci_receive( Nic_t *nic )
{
	Nic_Dm9pci_Ext_t	*ext;
	npkt_t			*npkt;
	npkt_t			*new;
	GenStats_t		*gstats;
	int				cidx;
	dm9pci_rx_desc	*rfd;
	int				pkt_len;
	int				pkt_cnt;
	uint32_t		status;
	int				ret;
	net_buf_t *buf;

	npkt_t *tx_Q_head = NULL, *tx_Q_tail = NULL;

#if DEBUG_RX
fprintf(dfd, "dm9pci_receive()\n\r");	
#endif
	ext		= (Nic_Dm9pci_Ext_t *)nic->ext;
	gstats	= &nic->nstats.gstats;

	npkt	= ext->npkt;
	pkt_cnt	= 0;
	ret		= 1;

#if DEBUG_RX
dump_rx_desc(ext);
#endif
	while ( ext->cr5_data & CSR5_RXCI )
	{
		cidx	= ext->rx_cidx;
		npkt	= (npkt_t *)ext->rx_pktq[cidx];
		rfd		= (dm9pci_rx_desc *)&ext->rx_desc[cidx];

		status	= rfd->rdes0;
		if( (status = rfd->rdes0) & RDES0_OWN )	/* card still owns this descriptor */
			break;

		/* advance consumer index */
		ext->rx_cidx = ( cidx + 1 ) % RX_DESC_NUM;
		pkt_len = ((status & RDES0_M_LENGTH) >> 16) - 4;	/* don't include CRC */
		ext->rx_pktq[cidx] = NULL;

		if(  pkt_len < ETH_MIN_PKT_LEN || pkt_len > ETH_MAX_PKT_LEN  )
		{
			gstats->rcv_error++;
			dm9pci_add_rfd( nic, npkt );
			continue;
		}

		if( new = ext->rx_free ) {
			ext->rx_free = new->next;
			new->next = NULL;
			ext->num_rx_free--;
		}
		else if ( !(new = dm9pci_alloc_npkt( nic, BUFFER_SIZE )) ) {
			/* can't allocate a new npkt for current Rx descriptor, drop the packet */
			gstats->rcv_error++;
			dm9pci_add_rfd( nic, npkt );
			continue;
		}

		dm9pci_add_rfd( nic, new );

		buf = TAILQ_FIRST(&npkt->buffers);
		buf->net_iov->iov_len = npkt->framelen = pkt_len;
        ext->txrx_count.rx_bytes += pkt_len;

		gstats->rcv_ok++;

		npkt->next = NULL;
		if(tx_Q_tail) {
			tx_Q_tail->next = npkt;
		}
		else {
			tx_Q_head = npkt;
		}
		tx_Q_tail = npkt;

		if( ++pkt_cnt > 10 ) {
			ret = 0;
			break;
		}
	}

	if(tx_Q_head) {
		pthread_mutex_unlock(&ext->mutex);
		if(tx_Q_head = ext->ion->tx_up_start(ext->reg_hdl, tx_Q_head, 0, 0, ext->cell, nic->lan, 0, nic)) {
			for( ; npkt = tx_Q_head; ) {
				tx_Q_head = npkt->next;
				dm9pci_receive_complete(npkt, nic, nic);
			}
		}
		pthread_mutex_lock(&ext->mutex);
	}

	return( ret );
}

int dm9pci_receive_complete( npkt_t *npkt, void *hdl, void *func_hdl)
{
	Nic_t			*nic;
	Nic_Dm9pci_Ext_t	*ext;

#if DEBUG_RX
fprintf(dfd, "dm9pci_receive_complete()\n\r");	
#endif
	nic		= (Nic_t *)hdl;
	ext	= (Nic_Dm9pci_Ext_t *)nic->ext;

	if( ext->num_rx_free < RX_DESC_NUM ) {
		npkt->ref_cnt = 1;
		npkt->req_complete = 0;
		npkt->flags = _NPKT_UP;

//		/* In case this came from dm9pci_advertise() which aligned to 32 bytes */
//		(int)npkt->buffers.tqh_first->net_iov->iov_base |= 0x1f;

		pthread_mutex_lock( &ext->mutex );
		npkt->next = ext->rx_free;
		ext->rx_free = npkt;
		ext->num_rx_free++;

		pthread_mutex_unlock( &ext->mutex );
		return ( 0 );
	}

	ion_free(npkt->org_data);
	ion_free(npkt);
	return ( 0 );
}

#if 0
/* The following 2 routines works for v0.01a, substiute if for improving reliability */
int dm9pci_receive( Nic_t *nic )
{
	Nic_Dm9pci_Ext_t	*ext;
	npkt_t			*npkt;
	net_buf_t		*nb;
	net_iov_t		*iov;
	int			pkt_len;
	int			pkt_cnt;
	int			ret;
	int			last_rx;
	int			entry;
	uint32_t		desc_status;

#if DEBUG_RX
fprintf(dfd, "dm9pci_receive()\n\r");	
#endif
	ext = (Nic_Dm9pci_Ext_t *)nic->ext;

	pkt_cnt = 0;
	ret = 1;
	last_rx = ext->cur_rx;

#if DEBUG_RX
dump_rx_desc(ext);
#endif
	while (!((desc_status = ext->rx_desc[ext->cur_rx].rdes0) & RDES0_OWN))
	{
		/* davance consumer index */
		pthread_mutex_unlock(&ext->mutex);
		entry = ext->cur_rx;
		ext->cur_rx++;
		ext->cur_rx %= RX_DESC_NUM;
		pthread_mutex_lock(&ext->mutex);

		pkt_cnt++;

		pkt_len = ((desc_status & RDES0_M_LENGTH) >> 16) - 4;
		if ((pkt_len < 60) || (pkt_len > 1514))
			continue;	

		npkt = (npkt_t *)ext->rx_desc[entry].rx_npkt_ptr;

		nb = npkt->buffers.tqh_first;
		iov = nb->net_iov;
		npkt->framelen = iov->iov_len = pkt_len;
//dump_npkt(npkt);

		npkt->next = NULL;

		pthread_mutex_unlock(&ext->mutex);
		ion_rx_packets(ext->reg_hdl, npkt, 0, 0, ext->cell, nic->lan, 0);
		npkt->framelen = iov->iov_len = BUFFER_SIZE;
//		dm9pci_receive_complete(npkt, nic, nic);
		ext->rx_desc[entry].rdes1 = RDES1_CE | MAX_FRAME_SIZE;
		ext->rx_desc[entry].rdes0 = RDES0_OWN;	
#if DEBUG_RX
//dump_rx_desc(ext);
#endif
		pthread_mutex_lock(&ext->mutex);

		ret = 0;
		if((pkt_cnt > 10) || (pkt_cnt >= RX_DESC_NUM))
			break;
	}

	return(ret);
}

int dm9pci_receive_complete( npkt_t *npkt, void *hdl, void *func_hdl)
{
	Nic_t			*nic;
	Nic_Dm9pci_Ext_t	*ext;
//net_buf_t		*nb;
//net_iov_t		*iov;

	nic		= (Nic_t *)hdl;
	ext		= (Nic_Dm9pci_Ext_t *)nic->ext;

#if DEBUG_RX
fprintf(dfd, "receive_complete()\n\r");
//fprintf(dfd, "CSR3 = %08x, cur_rx = %d\n\r",
//		IN32(ext, CSR3), ext->cur_rx);
#endif

//nb = npkt->buffers.tqh_first;
//iov = nb->net_iov;
//fprintf(dfd, "npkt->org_data=%08X, npkt->next=%08X, npkt->tot_iov=%d\n\r",
//	npkt->org_data,
//	npkt->next,
//	npkt->tot_iov);
//fprintf(dfd, "npkt->framelen=%d, iov->iov_len=%d\n\r",
//	npkt->framelen,
//	iov->iov_len);

//npkt->framelen = iov->iov_len = BUFFER_SIZE;

	ion_free(npkt->org_data);
	ion_free(npkt);

	return ( 0 );
}
#endif